from . import options
from .semantics import *

NAME = 'strings'
MUTATORS = ['str-constants']

class PassStringSimplifyConstant:
    """Replace a string constant by a shorter version."""
    def filter(self, node):
        return is_string_constant(node) and node != '""'
    def mutations(self, node):
        content = node[1:-1]
        return ['"{}"'.format(c) for c in ['', content[1:], content[:-1]]]
    def __str__(self):
        return 'simplify string constant'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'strings mutators')
    options.add_mutator_argument(argparser, 'str-constants', True, 'replaces constants by simpler ones')

def collect_mutators(args):
    res = []
    if args.mutator_strings:
        if args.mutator_str_constants:
            res.append(PassStringSimplifyConstant())
    return res



import copy
import progressbar

from . import mutator_options
from . import semantics

enabled_mutators = []
def collect_mutators(args):
    """Initializes the list of all active mutators."""
    global enabled_mutators
    enabled_mutators = mutator_options.collect_mutators(args)

def __mutate_node(node):
    """Apply all active mutators to the given node. Returns a list of all possible mutations."""
    res = []
    for m in enabled_mutators:
        if hasattr(m, 'filter') and not m.filter(node):
            continue
        res = res + list(map(lambda x: (str(m), x), m.mutations(node)))
    return res


def __generate_mutations(original, prg):
    """Generate mutations from the given original, updating the progress bar."""
    prg.update(prg.currval + 1)
    yield from __mutate_node(original)
    if isinstance(original, list):
        for i,o in enumerate(original):
            cand = copy.copy(original)
            for mutated in __generate_mutations(o, prg):
                cand[i] = mutated[1]
                yield (mutated[0], cand)

def __generate_mutations_bfs(original, prg):
    pass

def generate_mutations(original):
    """A generator that produces all possible mutations from the given original."""
    semantics.collect_information(original)
    s = semantics.node_count(original)
    widgets = [progressbar.Bar(), ' ', progressbar.Counter(), ' / ', str(s)]
    prg = progressbar.ProgressBar(maxval = s, widgets = widgets)
    prg.start()
    prg.update_interval = 1
    yield from __generate_mutations(original, prg)



import argparse

from . import mutators_arithmetic
from . import mutators_bitvectors
from . import mutators_boolean
from . import mutators_core
from . import mutators_smtlib
from . import mutators_strings

def disable(namespace, option):
    setattr(namespace, 'mutator_{}'.format(option.replace('-', '_')), False)
def disable_all(namespace, options):
    for o in options:
        disable(namespace, o)

class LetEliminationAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string = None):
        disable(namespace, mutators_arithmetic.NAME)
        disable(namespace, mutators_bitvectors.NAME)
        disable(namespace, mutators_boolean.NAME)
        disable_all(namespace, mutators_core.MUTATORS)
        disable(namespace, mutators_smtlib.NAME)
        disable(namespace, mutators_strings.NAME)
        setattr(namespace, 'mutator_let_elimination', True)
        setattr(namespace, 'mutator_let_substitution', True)

def collect_mutator_modes(argparser):
    argparser.add_argument('--mode-let-elimination', nargs = 0, action = LetEliminationAction, help = 'only eliminate let binders')

def collect_mutator_options(argparser):
    """Adds all options related to mutators to the given argument parser."""
    mutators_core.collect_mutator_options(argparser.add_argument_group('core mutator arguments'))
    mutators_boolean.collect_mutator_options(argparser.add_argument_group('boolean mutator arguments'))
    mutators_arithmetic.collect_mutator_options(argparser.add_argument_group('arithmetic mutator arguments'))
    mutators_bitvectors.collect_mutator_options(argparser.add_argument_group('bitvector mutator arguments'))
    mutators_smtlib.collect_mutator_options(argparser.add_argument_group('smtlib mutator arguments'))
    mutators_strings.collect_mutator_options(argparser.add_argument_group('strings mutator arguments'))

def collect_mutators(args):
    """Initializes the list of all active mutators."""
    res = []
    res += mutators_core.collect_mutators(args)
    res += mutators_boolean.collect_mutators(args)
    res += mutators_arithmetic.collect_mutators(args)
    res += mutators_bitvectors.collect_mutators(args)
    res += mutators_smtlib.collect_mutators(args)
    res += mutators_strings.collect_mutators(args)
    return res



import os

def version_from_git():
    import re
    import subprocess
    GIT_VERSION = subprocess.check_output(['git', 'describe', '--tags']).decode('utf8').strip()
    if re.match('^v[0-9.]+$', GIT_VERSION) is not None:
        return GIT_VERSION[1:]
    else:
        m = re.match('v([0-9]+)\\.([0-9]+)-([0-9]+)-g[a-z0-9]+', GIT_VERSION)
        if m is not None:
            return '{}.{}.dev{}'.format(m.group(1), int(m.group(2)) + 1, m.group(3))
    return GIT_VERSION

def version_from_package_metadata():
    try:
        from importlib import metadata
    except ImportError:
        import importlib_metadata as metadata
    return metadata.version('pydelta-smt')

if os.path.isdir('.git'):
    VERSION = version_from_git()
else:
    VERSION = version_from_package_metadata()

PACKAGE_NAME = 'pydelta-smt'
PROJECT_NAME = 'pyDelta'
DESCRIPTION = 'Delta debugger for SMT-LIB files'
AUTHOR = 'Gereon Kremer'
AUTHOR_EMAIL  = 'gereon.kremer@gmail.com'
REPOSITORY = 'https://github.com/nafur/pydelta'



from . import options
from .semantics import *

NAME = 'core'
MUTATORS = ['constants', 'erase-children', 'inline-functions', 'let-elimination', 'let-substitution', 'merge-children', 'replace-by-variable', 'sort-children', 'substitute-children']

class PassConstants:
    """Replaces any node by a constant."""
    def mutations(self, node):
        """Return :code:`get_constants(get_return_type(node))`."""
        res = get_constants(get_return_type(node))
        if node in res:
            return []
        return res
    def __str__(self):
        return 'substitute by a constant'

class PassEraseChildren:
    """Erases a single child of the given node."""
    def filter(self, node):
        return not is_leaf(node)
    def mutations(self, node):
        res = []
        for i in range(len(node)):
            tmp = node.copy()
            del tmp[i]
            res.append(tmp)
        return res
    def __str__(self):
        return 'erase child'

class PassInlineDefinedFuns:
    """Explicitly inlines a defined function."""
    def filter(self, node):
        return is_defined_function(node)
    def mutations(self, node):
        return [ get_defined_function(node)(node[1:]) ]
    def __str__(self):
        return 'inline defined functions'

class PassLetElimination:
    """Substitutes a let expression with its body."""
    def filter(self, node):
        return is_let(node)
    def mutations(self, node):
        return node[2]
    def __str__(self):
        return 'eliminate let binder'

class PassLetSubstitution:
    """Substitutes a variable bound by a let binder into the nested term."""
    def filter(self, node):
        return is_let(node)
    def mutations(self, node):
        res = []
        for var in node[1]:
            if contains(node[2], var[0]):
                subs = substitute(node[2], {var[0]: var[1]})
                res.append([node[0], node[1], subs])
        return res
    def __str__(self):
        return 'substitute variable into let body'

class PassMergeWithChildren:
    """Merges a node with one of its children. This is possible for n-ary operators like :code:`and` or :code:`+`."""
    def filter(self, node):
        return is_nary(node)
    def mutations(self, node):
        res = []
        for cid,child in enumerate(node):
            if has_name(child) and get_name(node) == get_name(child):
                res.append(node[:cid] + node[cid][1:] + node[cid+1:])
        return res
    def __str__(self):
        return 'merge with child'

class PassReplaceByVariable:
    """Replaces a node by a variable."""
    def filter(self, node):
        return node_count(node) > 1
    def mutations(self, node):
        return [v for v in get_variable_info().keys() if node_count(v) < node_count(node)]
    def __str__(self):
        return 'substitute by existing variable'

class PassSortChildren:
    """Sorts the children of a node."""
    def filter(self, node):
        return not is_leaf(node)
    def mutations(self, node):
        s = sorted(node, key = node_count)
        if s != node:
            return [s]
        return []
    def __str__(self):
        return 'sort children'

class PassSubstituteChildren:
    """Substitutes a node with one of its children."""
    def filter(self, node):
        return not is_leaf(node) and not is_let(node)
    def mutations(self, node):
        return node[1:]
    def __str__(self):
        return 'substitute with child'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'core mutators')
    options.add_mutator_argument(argparser, 'constants', True, 'replace by theory constants')
    options.add_mutator_argument(argparser, 'erase-children', True, 'erase individual children of nodes')
    options.add_mutator_argument(argparser, 'inline-functions', True, 'inline defined functions')
    options.add_mutator_argument(argparser, 'let-elimination', True, 'eliminate let bindings')
    options.add_mutator_argument(argparser, 'let-substitution', True, 'substitute bound variables in let bindings')
    options.add_mutator_argument(argparser, 'merge-children', True, 'merge children into nodes')
    options.add_mutator_argument(argparser, 'replace-by-variable', True, 'replace with existing variable')
    options.add_mutator_argument(argparser, 'sort-children', True, 'sort children of nodes')
    options.add_mutator_argument(argparser, 'substitute-children', True, 'substitute nodes with their children')

def collect_mutators(args):
    res = []
    if args.mutator_core:
        if args.mutator_constants:
            res.append(PassConstants())
        if args.mutator_erase_children:
            res.append(PassEraseChildren())
        if args.mutator_inline_functions:
            res.append(PassInlineDefinedFuns())
        if args.mutator_let_elimination:
            res.append(PassLetElimination())
        if args.mutator_let_substitution:
            res.append(PassLetSubstitution())
        if args.mutator_merge_children:
            res.append(PassMergeWithChildren())
        if args.mutator_replace_by_variable:
            res.append(PassReplaceByVariable())
        if args.mutator_sort_children:
            res.append(PassSortChildren())
        if args.mutator_substitute_children:
            res.append(PassSubstituteChildren())
    return res



from . import options
from .semantics import *

NAME = 'smtlib'
MUTATORS = ['check-sat-assuming']

class PassCheckSatAssuming:
    """Replaces a :code:`check-sat-assuming` by a regular :code:`check-sat`."""
    def filter(self, node):
        return has_name(node) and get_name(node) == 'check-sat-assuming'
    def mutations(self, node):
        return [['check-sat']]
    def __str__(self):
        return 'substitute check-sat-assuming by check-sat'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'smtlib mutators')
    options.add_mutator_argument(argparser, 'check-sat-assuming', True, 'replace check-sat-assuming by check-sat')

def collect_mutators(args):
    res = []
    if args.mutator_smtlib:
        if args.mutator_check_sat_assuming:
            res.append(PassCheckSatAssuming())
    return res



import setuptools

from pydelta import info

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name = info.PACKAGE_NAME,
    version = info.VERSION,
    description = info.DESCRIPTION,
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = info.REPOSITORY,
    author = info.AUTHOR,
    author_email = info.AUTHOR_EMAIL,
    license = 'MIT',
    packages = setuptools.find_packages(),
    zip_safe = False,
    scripts = [
        'bin/pydelta'
    ],
    install_requires = [
        'progressbar>=2.5',
        'sphinx-rtd-theme>=0.4.3',
        'importlib-metadata>=1.7 ; python_version<"3.8"',
    ]
)
